// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2011 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Threading;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapOpenedMailboxInfo {
    private const int defaultPollingInterval = 10 * 60 * 1000; // 10 mins
    private const int maximumPollingInterval = 30 * 60 * 1000; // 30 mins

    /*
     * WaitForMessageArrival
     */
    public ImapMessageInfoList WaitForMessageArrival(TimeSpan timeout)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   (int)timeout.TotalMilliseconds,
                                   ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan timeout,
                                                     ImapMessageFetchAttributeOptions options)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   (int)timeout.TotalMilliseconds,
                                   options);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsTimeout)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   millisecondsTimeout,
                                   ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsTimeout,
                                                     ImapMessageFetchAttributeOptions options)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   millisecondsTimeout,
                                   options);
    }

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan pollingInterval,
                                                     TimeSpan timeout)
    {
      return WaitForMessageArrival((int)pollingInterval.TotalMilliseconds,
                                   (int)timeout.TotalMilliseconds,
                                   ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan pollingInterval,
                                                     TimeSpan timeout,
                                                     ImapMessageFetchAttributeOptions options)
    {
      return WaitForMessageArrival((int)pollingInterval.TotalMilliseconds,
                                   (int)timeout.TotalMilliseconds,
                                   options);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsPollingInterval,
                                                     int millisecondsTimeout)
    {
      return WaitForMessageArrival(millisecondsPollingInterval,
                                   millisecondsTimeout,
                                   ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsPollingInterval,
                                                     int millisecondsTimeout,
                                                     ImapMessageFetchAttributeOptions options)
    {
      if (millisecondsTimeout < -1)
        throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, "millisecondsTimeout", millisecondsTimeout);

      var asyncResult = BeginWaitForMessageArrival(millisecondsPollingInterval, options, null, null);

      asyncResult.AsyncWaitHandle.WaitOne(millisecondsTimeout, false);

      return EndWaitForMessageArrival(asyncResult);
    }

    /*
     * BeginWaitForMessageArrival
     */
    public IAsyncResult BeginWaitForMessageArrival(object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival(defaultPollingInterval,
                                        ImapMessageFetchAttributeOptions.Default,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(TimeSpan pollingInterval,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival((int)pollingInterval.Milliseconds,
                                        ImapMessageFetchAttributeOptions.Default,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(int millisecondsPollingInterval,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival(millisecondsPollingInterval,
                                        ImapMessageFetchAttributeOptions.Default,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(TimeSpan pollingInterval,
                                                   ImapMessageFetchAttributeOptions options,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival((int)pollingInterval.TotalMilliseconds,
                                        options,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(int millisecondsPollingInterval,
                                                   ImapMessageFetchAttributeOptions options,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      if (millisecondsPollingInterval <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("millisecondsPollingInterval", millisecondsPollingInterval);
      if (waitForMessageArrivalAsyncResult != null)
        throw new InvalidOperationException("another operation proceeding");

      var isIdleCapable = Client.ServerCapabilities.Contains(ImapCapability.Idle);

      if (!isIdleCapable && maximumPollingInterval <= millisecondsPollingInterval)
        throw new ArgumentOutOfRangeException("millisecondsPollingInterval",
                                              millisecondsPollingInterval,
                                              string.Format("maximum interval is {0} but requested was {1}", maximumPollingInterval, millisecondsPollingInterval));

      var proc = isIdleCapable
        ? new WaitForMessageArrivalProc(WaitForMessageArrivalIdle)
        : new WaitForMessageArrivalProc(WaitForMessageArrivalNoOp);

      waitForMessageArrivalAsyncResult = new WaitForMessageArrivalAsyncResult(millisecondsPollingInterval,
                                                                              Mailbox,
                                                                              options,
                                                                              proc,
                                                                              asyncState,
                                                                              asyncCallback);

      waitForMessageArrivalAsyncResult.BeginProc();

      return waitForMessageArrivalAsyncResult;
    }

    public ImapMessageInfoList EndWaitForMessageArrival(IAsyncResult asyncResult)
    {
      if (asyncResult == null)
        throw new ArgumentNullException("asyncResult");
      if (asyncResult != waitForMessageArrivalAsyncResult)
        throw ExceptionUtils.CreateArgumentMustBeValidIAsyncResult("asyncResult");

      try {
        waitForMessageArrivalAsyncResult.EndProc();

        return GetArrivalMessages(waitForMessageArrivalAsyncResult);
      }
      finally {
        waitForMessageArrivalAsyncResult = null;
      }
    }

    private ImapMessageInfoList GetArrivalMessages(WaitForMessageArrivalContext context)
    {
      try {
        return new ImapMessageInfoList(this, context.FetchOptions, new ArrivalMessageQuery(context));
      }
      finally {
        context.Dispose();
      }
    }

    private class ArrivalMessageQuery : IImapMessageQuery {
      public ArrivalMessageQuery(WaitForMessageArrivalContext context)
      {
        sequenceSet = (!context.MessageArrived || context.Mailbox.ExistsMessage <= context.CurrentMessageCount)
          ? ImapSequenceSet.CreateSet(new long[0])
          : ImapSequenceSet.CreateRangeSet(context.CurrentMessageCount + 1, context.Mailbox.ExistsMessage);
      }

      public ImapSequenceSet GetSequenceOrUidSet(ImapOpenedMailboxInfo mailbox)
      {
        return sequenceSet;
      }

      private ImapSequenceSet sequenceSet;
    }

    private delegate bool WaitForMessageArrivalProc(WaitForMessageArrivalContext context);
    private WaitForMessageArrivalAsyncResult waitForMessageArrivalAsyncResult = null;

    private class WaitForMessageArrivalContext : IDisposable {
      public bool MessageArrived {
        get; protected set;
      }

      public WaitHandle AbortEvent {
        get { return abortEvent; }
      }

      public long CurrentMessageCount {
        get; set;
      }

      public int PollingInterval {
        get; private set;
      }

      public ImapMailbox Mailbox {
        get; private set;
      }

      public ImapMessageFetchAttributeOptions FetchOptions {
        get; private set;
      }

      public WaitForMessageArrivalContext(int pollingInterval,
                                          ImapMailbox mailbox,
                                          ImapMessageFetchAttributeOptions fetchOptions)
      {
        this.PollingInterval = pollingInterval;
        this.Mailbox = mailbox;
        this.CurrentMessageCount = mailbox.ExistsMessage;
        this.FetchOptions = fetchOptions;
      }

      public void Dispose()
      {
        if (abortEvent != null) {
          abortEvent.Close();
          abortEvent = null;
        }
      }

      protected void Abort()
      {
        abortEvent.Set();
      }

      private ManualResetEvent abortEvent = new ManualResetEvent(false);
    }

    private class WaitForMessageArrivalAsyncResult :
      WaitForMessageArrivalContext,
      IAsyncResult
    {
      public object AsyncState {
        get; private set;
      }

      public WaitHandle AsyncWaitHandle {
        get { return procAsyncResult.AsyncWaitHandle; }
      }

      public bool CompletedSynchronously {
        get { return false; }
      }

      public bool IsCompleted {
        get { return procAsyncResult.AsyncWaitHandle.WaitOne(0, false); }
      }

      private AsyncCallback asyncCallback;
      private WaitForMessageArrivalProc proc;
      private IAsyncResult procAsyncResult;

      public WaitForMessageArrivalAsyncResult(int pollingInterval,
                                              ImapMailbox mailbox,
                                              ImapMessageFetchAttributeOptions fetchOptions,
                                              WaitForMessageArrivalProc proc,
                                              object asyncState,
                                              AsyncCallback asyncCallback)
        : base(pollingInterval, mailbox, fetchOptions)
      {
        this.proc = proc;
        this.AsyncState = asyncState;
        this.asyncCallback = asyncCallback;
      }

      public void BeginProc()
      {
        procAsyncResult = proc.BeginInvoke(this, null, null);
      }

      public void EndProc()
      {
        Abort();

        MessageArrived = proc.EndInvoke(procAsyncResult);

        if (MessageArrived && asyncCallback != null)
          asyncCallback(this);
      }
    }

    private bool WaitForMessageArrivalIdle(WaitForMessageArrivalContext context)
    {
      var waitHandles = new WaitHandle[] {context.AbortEvent, null};

      for (;;) {
        if (context.AbortEvent.WaitOne(0, false))
          return false; // aborted

        var idleAsyncResult = Client.Session.BeginIdle(context, delegate(object idleState, ImapUpdatedStatus updatedStatus) {
          var ctx = idleState as WaitForMessageArrivalAsyncResult;

          if (updatedStatus.Expunge.HasValue)
            ctx.CurrentMessageCount--;
          else if (updatedStatus.Exists.HasValue && ctx.CurrentMessageCount < updatedStatus.Exists.Value)
            return false; // DONE

          return true; // keep IDLEing
        });

        waitHandles[1] = idleAsyncResult.AsyncWaitHandle;

        var index = WaitHandle.WaitAny(waitHandles, 180 * 1000, false); // send DONE 3 minute after IDLE started

        ProcessResult(Client.Session.EndIdle(idleAsyncResult));

        switch (index) {
          case 0: // aborted
            return false;

          case 1: // IDLE completed
            return true;

          case WaitHandle.WaitTimeout:
          default:
            continue; // re-IDLE
        }
      }
    }

    private bool WaitForMessageArrivalNoOp(WaitForMessageArrivalContext context)
    {
      for (;;) {
        if (context.AbortEvent.WaitOne(context.PollingInterval, false))
          return false; // aborted

        var result = ProcessResult(Client.Session.NoOp());

        foreach (var response in result.ReceivedResponses) {
          var dataResponse = response as ImapDataResponse;

          if (dataResponse == null)
            continue;

          if (dataResponse.Type == ImapDataResponseType.Expunge)
            context.CurrentMessageCount--;
        }

        if (context.CurrentMessageCount < context.Mailbox.ExistsMessage)
          return true; // new message arrival
      }
    }
  }
}

